# pygpt64/__init__.py

# Definisci la versione del pacchetto
__version__ = '0.1'

# Importa i simboli necessari dai moduli interni
#from .main import *
from .main_class import *
#from .assistants import *

# Se desideri esporre specifiche classi o funzioni da altri moduli, puoi farlo qui.
# Ad esempio:
# from .module1 import Class1, function1
# from .module2 import Class2, function2
